var request = require('request');
var url = require('url');
var winston = require('winston');
var xml2jsParser = require('xml2js').parseString;
var redis = require('redis');

var config;
var aaa;
var callbackReport;
var logger;
var redisClient;

var maxCheckRetry = 20;
var delayBeforeCheckRetry = 30 * 1000;

process.env.NODE_TLS_REJECT_UNAUTHORIZED = '0';

function createRedisClient(host, port) {
    try {
        redisClient = redis.createClient(port, host);
    } catch(err) {
        logger.warn("Error creating redis client to " + host + ':' + port);
    }
}

function start(_config, _callbackReport, options) {
    config = _config;
    callbackReport = _callbackReport;

    if (options && options.aaa) {
            aaa = options.aaa;
    }

    if (options && options.logger) {
        logger = options.logger;
    } else {
        logger = new winston.Logger({
            transports: [
              new (winston.transports.Console)()
            ]
        });
    }

    createRedisClient(config.globals.redis_host, config.globals.redis_port);
}

function createRequestOptions(methodName, task) {
    var partnerUrl = url.parse(config.h2h_out.partner);
    var product = prepareRemoteProductCode(task.remoteProduct);

    var destination = task.destination;
    if (methodName == 'CHECKING') {
        destination = paddingDestination(destination);
    }

    var options = {
        url: config.h2h_out.partner,
        qs: {
            request: methodName + '*'
                + task.requestId + '*'
                + product.product + '*'
                + product.productDetail + '*'
                + destination + '*'
                + product.nominal + '*'
                + '0*'
                + config.h2h_out.noid + '*'
                + config.h2h_out.userid + '*'
                + config.h2h_out.password
        }
    };

    return options;
}

function topupCheck(task, _retry) {
    // disable topupCheck sampai dapat konfirmasi perubahan spesifikasi nomor tujuan 13 digit pading left 0
    return;

    var retry = _retry;

    if (retry === null || retry === undefined) {
        retry = maxCheckRetry + 1;
    }

    retry = retry - 1;

    if (retry < 0) {
        var message = 'Max retry check transaction retry exceeded';
        var response = {
            raw: message,
            parsed: {
                MESSAGE: message
            }
        }
        aaa.pushResponseToMongoDb(task, response, '68');
        callbackReport(task.requestId, '68', message);
        return;
    }

    requestToPartner('CHECKING', task, retry);
}

function topupRequest(task) {
    if (!aaa.isTodayTrx(task)) {
        logger.warn('Maaf, transaksi beda hari tidak dapat dilakukan');
        callbackReport(task.requestId, '68', 'Maaf, transaksi beda hari tidak dapat dilakukan');
        return;
    }

    getTaskForDupeCheck(task, function(err, dupeTask) {
        if (dupeTask) {
            logger.info('Duplicate task detected, requesting topupCheck', {task: task});
            topupCheck(task);
        } else {
            saveTaskForDupeCheck(task);
            requestToPartner('PURCHASE', task);
        }
    });

}

function requestToPartner(methodName, task, retry) {
    aaa.insertTaskToMongoDb(task);

    var options = createRequestOptions(methodName, task);

    logger.info('Requesting to partner', {requestOption: options});
    request(options, function(error, response, body) {
        if (error) {
            var msg = 'Error requesting to partner. ' + error;
            var rc = '68';

            try {
                if (methodName == 'PURCHASE') {
                    if ((error.connect === true || error.syscall == "connect") && (error.code === 'ETIMEDOUT' || error.code === 'ECONNREFUSED')) {
                        rc = '40';
                    }
                }
            }
            catch(e) {
                logger.warn('Parsing error exception on requestToPartner()');
            }

            logger.warn(msg, {error: error});
            callbackReport(task.requestId, rc, msg);

            var _response = {
                raw: JSON.stringify({msg: msg, error: error}),
                parsed: {
                    MESSAGE: msg,
                    error: error
                }
            }
            aaa.pushResponseToMongoDb(task, _response, rc);

            if (rc == '68') {
                setTimeout(
                    topupCheck,
                    delayBeforeCheckRetry,
                    task,
                    retry
                );
            }

            return;
        }

        if (response.statusCode != 200) {
            var message = 'Partner response with http status code other that 200 (' + response.statusCode + ')';
            var rc = '68';

            if (!retry && response.statusCode == 404) {
                rc = '91';
            }

            logger.warn(message);
            callbackReport(task.requestId, rc, message);

            var _response = {
                raw: message,
                parsed: {
                    MESSAGE: message,
                    responseHttpStatus: response.statusCode,
                    responseBody: body,
                }
            }
            aaa.pushResponseToMongoDb(task, _response, rc);

            if (rc == '68') {
                setTimeout(
                    topupCheck,
                    delayBeforeCheckRetry,
                    task,
                    retry
                );
            }

            return;
        }

        logger.verbose('Got respose', {rawBody: body});
        parseResponse(body, task);
    });
}

function getSn(response) {
    try {
        var sn = response.xml.ket1[0];
        return sn;
    }
    catch (err) {
        return;
    }
}

function getHarga(response) {
    try {
        var harga = response.xml.amount[0];
        return harga;
    }
    catch (err) {
        return;
    }
}

function getSaldo(response) {
    try {
        var saldo = response.xml.saldo[0];
        return saldo;
    }
    catch (err) {
        return;
    }
}

function parseResponse(body, task) {
    xml2jsParser(body, function(err, response) {
        if (err) {
            logger.warn('Error parsing XML', {error: err, task: task, responseBody: body});

            var message = 'Error parsing XML. ' + err + '. ' + body;

            var _response = {
                raw: body,
                parsed: {
                    MESSAGE: message
                }
            }
            aaa.pushResponseToMongoDb(task, _response, '68');

            callbackReport(task.requestId, '68', message);
            return;
        }

        logger.info('Got response', {response: response});

        var responseCode;
        var message;

        try {
            responseCode = response.xml.response_code[0];
            message = response.xml.response_message[0];
        }
        catch(errGetParam) {
            logger.warn('Exception saat parsing hasil', {error: errGetParam, task: task, responseBody: body});

            var _response = {
                raw: body,
                parsed: {
                    MESSAGE: 'Exception saat parsing hasil. ' + errGetParam,
                    xml: response.xml
                }
            }
            aaa.pushResponseToMongoDb(task, _response, '68');

            callbackReport(task.requestId, '68', 'Exception saat parsing hasil. ' +  errGetParam);
            return;
        }

        var st24rc;
        if (parseInt(responseCode) == 0) {
            st24rc = '00';
        }
        else if (parseInt(responseCode) == 99) {
            st24rc = '68'
        }
        else if (parseInt(responseCode) == 201) {
            st24rc = '68'
        }
        else if (message.indexOf('Transaksi sedang dalam proses') >= 0) {
            st24rc = '68';
        }
        else {
            st24rc = '40';
        }

        var st24message = message;
        if (responseCode) {
            st24message = responseCode + ' ' + st24message;
        }

        if (st24rc == '00') {
            var sn = getSn(response);

            if (sn) {
                st24message = 'SN=' + sn + ';' + st24message + '. SN=' + sn;
            }
        }

        var harga = getHarga(response);
        if (harga) {
            st24message = st24message + '. Harga ' + harga;
        }

        var saldo = getSaldo(response);
        if (saldo) {
            st24message = st24message + '. Saldo ' + saldo;
            aaa.updateBalance(saldo);
        }

        var _response = {
            raw: body,
            parsed: {
                MESSAGE: st24message,
                xml: response.xml
            }
        }

        aaa.pushResponseToMongoDb(task, _response, st24rc);

        callbackReport(task.requestId, st24rc, st24message);
    });
}

function prepareRemoteProductCode(remoteProduct) {
    var product = remoteProduct.split(',');

    if (product.length != 3) {
        return;
    }

    return {
        product: product[0],
        productDetail: product[1],
        nominal: product[2]
    }
}

function paddingDestination(destination, _width) {
    var width = _width;
    if (!width) {
        width = 13;
    }

    if (destination.length > width) {
        return destination;
    }

    var padder = "000000000000000000000000000";

    return (padder + destination).slice(-1 * width);
}

function getTaskKeyForDupeCheck(task) {
    return config.globals.gateway_name + '.dupecheck.reqId:' + task.requestId;
}

function saveTaskForDupeCheck(task, cb) {
    var key = getTaskKeyForDupeCheck(task);

    logger.verbose('Saving ' + key + ' to redis for dupecheck protection');

    redisClient.set(key, JSON.stringify(task), function() {
        redisClient.expire(key, 3600*24*7);
        if (cb) {
            cb(null, task);
        }
    });
}

function getTaskForDupeCheck(task, cb) {
    var key = getTaskKeyForDupeCheck(task);

    redisClient.get(key, function(err, result) {
        if (err) {
            cb(err, null);
            return;
        }

        var taskOnRedis = null;
        try {
            taskOnRedis = JSON.parse(result);
        }
        catch(err1) {
            cb(err1, null);
            return;
        }

        cb(null, taskOnRedis);
    });
}

exports.start = start;
exports.topupRequest = topupRequest;
exports.prepareRemoteProductCode = prepareRemoteProductCode;
exports.paddingDestination = paddingDestination;
